/****************************************************************/
/*  
     ***   ***  ***  *** ***     *** ***   ******  ****  ***
     ***   ***  ***  *** ***     *** **   ***  *** ***** ***
     ***   ***  ***  *** ***     *****    ******** *** *****
      *** ***   ***  *** ******* *** **   ***  *** ***  ****
       *****     *****   ******* *** ***  ***  *** ***   ***
	   
	 Vulkan Graphics API: in 20 Minutes (Coffee Break Series)
	 Beginners Taste

     Minimalistic examples to build upon and experiment with 
	 to develop an understanding of the underlying API

	   
/*
/****************************************************************/
/*
     About:
	 Minimal Vulkan demo (i.e., stripped down to the basics).
	 Remember to compile and include the shader
	 files (vertex and pixel shaders).
	 e.g., run the compilershaders.bat

	 Extension Chapters 0X - Computer Shader (with Graphical Output)

	 - single queue
	 - basic bounding points
	 - no uniforms or graphical transforms (i.e., 1x1x1 unit view box)

	 Further ideas:
	 - mouse presses (add noise)
	 - add rotation (orientation and angular rotation)
	 - instancing (draw small shapes instead of points)
	 - instancing to draw little arrow for the velocity
	 - limit to 2D but full 3D possible
	 - texture ids (different particles with different textures)
	 - billboarding
	 - integrate into a virtual world (fires, dust, ..)
	 - organise into a reusable class
	 - extend to demonstrate metaballs (sticky blobs)
*/




#pragma comment(lib, "C:\\VulkanSDK\\1.0.17.0\\Bin32\\vulkan-1.lib")


#define _CRT_SECURE_NO_WARNINGS // fopen

#pragma comment(linker, "/subsystem:windows")
#include <windows.h>

#include "vulkan.h"

#include <stdio.h>
#include <stdlib.h>

using namespace std;
#include <vector>
#include <array>


#define VERTEX_BUFFER_BIND_ID     0

#include "helpers.h"
#include "vulkanbuffer.h"


#define PARTICLE_COUNT 30 
//#define PARTICLE_COUNT 256 
//#define PARTICLE_COUNT 1024 
//#define PARTICLE_COUNT 4096
//#define PARTICLE_COUNT 65568
//#define PARTICLE_COUNT 1048576
//#define PARTICLE_COUNT 16777216



/**************************************************************/
/*
  Main class that does all the work (VulkanDemo)


*/
/**************************************************************/




VkCoffeeDevice *						coffeeDevice		= NULL;
VkCoffeeSwapChain*						coffeeSwapChain		= NULL;
VkFormat								colorformat			= VK_FORMAT_B8G8R8A8_UNORM;
VkFormat								depthFormat			= VK_FORMAT_D32_SFLOAT;

class VulkanDemo
{
public:

	// `alignment' - why not a `vec3'?  When declaring UBOs/SSBOs, pretend 
	// all 3-element vector and matrix types don't exist. Pretend that the only types 
	// are scalars, 2, and 4 element vectors (and matrices).
	// Otherwise, you'll get strange issues - alignment/operations.
	struct Particle
	{
		float pos [4];
		float vel [4];
	};

	VkDescriptorPool					descriptorPool		= VK_NULL_HANDLE;
	VkDescriptorSetLayout				descriptorSetLayout	= VK_NULL_HANDLE;
	VkDescriptorSet						descriptorSet		= VK_NULL_HANDLE;			
	VkPipelineLayout					pipelineLayout		= VK_NULL_HANDLE;		
	VkPipeline							pipeline			= VK_NULL_HANDLE;		

	VkCommandPool						cmdPool				= 0;
	std::vector<VkCommandBuffer>		clearCmdBuffers;
	std::vector<VkFramebuffer>			frameBuffers;
	uint32_t							currentBuffer		= 0;
	VkRenderPass						renderPass			= 0;
	VkShaderModule						shaderModuleVS		= 0;
	VkShaderModule						shaderModulePS		= 0;
	VkShaderModule						shaderModuleCS		= 0;
									 
	const char*							title				= "Rendering with Vulkan (Compute Shader)";
	const char*							name				= "Vulkan Graphics API: in 20 Minutes (Coffee Break Series)";
	HWND								windowHWND			= 0;
	HINSTANCE							windowInstance		= 0;
	uint32_t							width				= 1280;
	uint32_t							height				= 720;

	struct 
	{
		VkImage					image;
		VkDeviceMemory			mem;
		VkImageView				view;
	} depthStencil;

	struct 
	{
		VkPipelineVertexInputStateCreateInfo			inputState;
		std::vector<VkVertexInputBindingDescription>	bindingDescriptions;
		std::vector<VkVertexInputAttributeDescription>	attributeDescriptions;
	} vertices;

	// Resources for the compute part of the example
	struct 
	{
		// (Shader) storage buffer object containing the particles
		Buffer								storageBuffer;
		// Command buffer storing the dispatch commands and barriers						
		VkCommandBuffer						commandBuffer		= 0;				
		// Compute shader binding layout
		VkDescriptorSetLayout				descriptorSetLayout	= 0;	
		// Compute shader bindings
		VkDescriptorSet						descriptorSet		= 0;				
		// Layout of the compute pipeline
		VkPipelineLayout					pipelineLayout		= 0;
		// Compute pipeline for updating particle positions			
		VkPipeline pipeline;						
	} compute;



	



	// Todo - release resources/assets/memory
	void ReleaseVulkan()
	{
	}
	


	void CreateCommandBuffers()
	{
		// Create command buffer for each swap chain image and reuse for rendering
		DBG_ASSERT(clearCmdBuffers.size()==0);
		clearCmdBuffers.resize(coffeeSwapChain->imageCount);

		VkCommandBufferAllocateInfo cmdBufAllocateInfo = {};
		cmdBufAllocateInfo.sType              = VK_STRUCTURE_TYPE_COMMAND_BUFFER_ALLOCATE_INFO;
		cmdBufAllocateInfo.commandPool        = cmdPool;
		cmdBufAllocateInfo.level              = VK_COMMAND_BUFFER_LEVEL_PRIMARY;
		cmdBufAllocateInfo.commandBufferCount = static_cast<uint32_t>(clearCmdBuffers.size());

		VK_CHECK_RESULT(vkAllocateCommandBuffers(coffeeDevice->device, &cmdBufAllocateInfo, clearCmdBuffers.data()));
	}

	void BuildComputeCommandBuffer()
	{
		VkCommandBufferBeginInfo cmdBufInfo {};
		cmdBufInfo.sType = VK_STRUCTURE_TYPE_COMMAND_BUFFER_BEGIN_INFO;
		
		VK_CHECK_RESULT(vkBeginCommandBuffer(compute.commandBuffer, &cmdBufInfo));

		// Compute particle movement

		// Add memory barrier to ensure that the (graphics) vertex shader has fetched attributes before compute starts to write to the buffer
		VkBufferMemoryBarrier bufferBarrier {};
		bufferBarrier.sType					= VK_STRUCTURE_TYPE_BUFFER_MEMORY_BARRIER;
		bufferBarrier.srcQueueFamilyIndex	= VK_QUEUE_FAMILY_IGNORED;
		bufferBarrier.dstQueueFamilyIndex	= VK_QUEUE_FAMILY_IGNORED;
		bufferBarrier.buffer				= compute.storageBuffer.buffer;
		bufferBarrier.size					= compute.storageBuffer.descriptor.range;
		// Vertex shader invocations have finished reading from the buffer
		bufferBarrier.srcAccessMask			= VK_ACCESS_VERTEX_ATTRIBUTE_READ_BIT;						
		// Compute shader wants to write to the buffer
		bufferBarrier.dstAccessMask			= VK_ACCESS_SHADER_WRITE_BIT;								
		// Compute and graphics queue may have different queue families (see VulkanDevice::createLogicalDevice)
		// For the barrier to work across different queues, we need to set their family indices
		bufferBarrier.srcQueueFamilyIndex	= coffeeDevice->queueFamilyIndices.graphics;			
		bufferBarrier.dstQueueFamilyIndex	= coffeeDevice->queueFamilyIndices.compute;			

		
		vkCmdPipelineBarrier(
			compute.commandBuffer,
			VK_PIPELINE_STAGE_VERTEX_SHADER_BIT,
			VK_PIPELINE_STAGE_COMPUTE_SHADER_BIT,
			0,// No flags
			0, nullptr,
			1, &bufferBarrier,
			0, nullptr);

		vkCmdBindPipeline(compute.commandBuffer, VK_PIPELINE_BIND_POINT_COMPUTE, compute.pipeline);
		vkCmdBindDescriptorSets(compute.commandBuffer, VK_PIPELINE_BIND_POINT_COMPUTE, compute.pipelineLayout, 0, 1, &compute.descriptorSet, 0, 0);

		// Dispatch the compute job
		vkCmdDispatch(compute.commandBuffer, PARTICLE_COUNT, 1, 1);


		// Add memory barrier to ensure that compute shader has finished writing to the buffer
		// Without this the (rendering) vertex shader may display incomplete results (partial data from last frame) 
		// Compute shader has finished writes to the buffer
		bufferBarrier.srcAccessMask = VK_ACCESS_SHADER_WRITE_BIT;								
		// Vertex shader invocations want to read from the buffer
		bufferBarrier.dstAccessMask = VK_ACCESS_VERTEX_ATTRIBUTE_READ_BIT;						
		bufferBarrier.buffer		= compute.storageBuffer.buffer;
		bufferBarrier.size			= compute.storageBuffer.descriptor.range;
		// Compute and graphics queue may have different queue families (see VulkanDevice::createLogicalDevice)
		// For the barrier to work across different queues, we need to set their family indices
		bufferBarrier.srcQueueFamilyIndex = coffeeDevice->queueFamilyIndices.compute;			
		bufferBarrier.dstQueueFamilyIndex = coffeeDevice->queueFamilyIndices.graphics;

		//DBG_ASSERT(0);

		vkCmdPipelineBarrier(
			compute.commandBuffer,
			VK_PIPELINE_STAGE_COMPUTE_SHADER_BIT,
			VK_PIPELINE_STAGE_VERTEX_SHADER_BIT,
			0, // No flags
			0, nullptr,
			1, &bufferBarrier,
			0, nullptr);

		vkEndCommandBuffer(compute.commandBuffer);
	}

	

	static
	float RandomFloat(float lo, float hi) 
	{
		float random = ((float) rand()) / (float) RAND_MAX;
		float diff   = hi - lo;
		float r      = random * diff;
		return lo + r;
	}


	// Setup and fill the compute shader storage buffers containing the particles
	void CreateComputeStorageBuffers()
	{
		// Initial particle positions/velocities
		std::vector<Particle> particleBuffer(PARTICLE_COUNT);
		for (int i=0; i<PARTICLE_COUNT; ++i)
		{
			float speed = 0.001f;

			particleBuffer[i].pos[0] = RandomFloat(-1,1);
			particleBuffer[i].pos[1] = RandomFloat(-1,1);
			particleBuffer[i].pos[2] = 0;
			particleBuffer[i].pos[3] = 1;

			particleBuffer[i].vel[0] = RandomFloat(-1,1) * speed;
			particleBuffer[i].vel[1] = RandomFloat(-1,1) * speed;
			particleBuffer[i].vel[2] = 0;
			particleBuffer[i].vel[3] = 0;
		}

		VkDeviceSize storageBufferSize = particleBuffer.size() * sizeof(Particle);

		// Staging - copy the memory once over to the device
		Buffer stagingBuffer;

		Buffer::CreateBuffer(
			VK_BUFFER_USAGE_TRANSFER_SRC_BIT,
			VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT | VK_MEMORY_PROPERTY_HOST_COHERENT_BIT,
			&stagingBuffer,
			storageBufferSize,
			particleBuffer.data());

		Buffer::CreateBuffer(
			// Buffer will be used as `both' a storage buffer for the compute pipeline and as 
			// a vertex buffer in the graphics pipeline
			VK_BUFFER_USAGE_VERTEX_BUFFER_BIT | VK_BUFFER_USAGE_STORAGE_BUFFER_BIT | VK_BUFFER_USAGE_TRANSFER_DST_BIT,
			VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT,
			&compute.storageBuffer,
			storageBufferSize);


		VkCommandBufferAllocateInfo cmdBufAllocateInfo {};
		cmdBufAllocateInfo.sType				= VK_STRUCTURE_TYPE_COMMAND_BUFFER_ALLOCATE_INFO;
		cmdBufAllocateInfo.commandPool			= cmdPool;
		cmdBufAllocateInfo.level				= VK_COMMAND_BUFFER_LEVEL_PRIMARY;
		cmdBufAllocateInfo.commandBufferCount	= 1;

		
		VkCommandBuffer copyCmd;
		VK_CHECK_RESULT(vkAllocateCommandBuffers(coffeeDevice->device, &cmdBufAllocateInfo, &copyCmd));

		VkCommandBufferBeginInfo cmdBufInfo {};
		cmdBufInfo.sType = VK_STRUCTURE_TYPE_COMMAND_BUFFER_BEGIN_INFO;
		VK_CHECK_RESULT(vkBeginCommandBuffer(copyCmd, &cmdBufInfo));

		// Copy to staging buffer
		VkBufferCopy copyRegion = {};
		copyRegion.size = storageBufferSize;
		vkCmdCopyBuffer(copyCmd, stagingBuffer.buffer, compute.storageBuffer.buffer, 1, &copyRegion);
		
		VK_CHECK_RESULT(vkEndCommandBuffer(copyCmd));

		VkSubmitInfo submitInfo				= {};
		submitInfo.sType					= VK_STRUCTURE_TYPE_SUBMIT_INFO;
		submitInfo.commandBufferCount		= 1;
		submitInfo.pCommandBuffers			= &copyCmd;

		VK_CHECK_RESULT(vkQueueSubmit(coffeeDevice->queue, 1, &submitInfo, VK_NULL_HANDLE));
		VK_CHECK_RESULT(vkQueueWaitIdle(coffeeDevice->queue));

		vkFreeCommandBuffers(coffeeDevice->device, cmdPool, 1, &copyCmd);
	
		// free stagingbuffer resources
		if (stagingBuffer.buffer)
		{
			vkDestroyBuffer(coffeeDevice->device, stagingBuffer.buffer, nullptr);
		}
		if (stagingBuffer.memory)
		{
			vkFreeMemory(coffeeDevice->device, stagingBuffer.memory, nullptr);
		}


		VkVertexInputBindingDescription vInputBindDescription0 {};
		vInputBindDescription0.binding		= VERTEX_BUFFER_BIND_ID;
		vInputBindDescription0.stride		= sizeof(Particle);
		vInputBindDescription0.inputRate	= VK_VERTEX_INPUT_RATE_VERTEX;

		// Binding description
		DBG_ASSERT(vertices.bindingDescriptions.size()==0);
		vertices.bindingDescriptions.push_back( vInputBindDescription0 );


		VkVertexInputAttributeDescription vInputAttribDescription0 {};
		vInputAttribDescription0.location	= 0;
		vInputAttribDescription0.binding	= VERTEX_BUFFER_BIND_ID;
		vInputAttribDescription0.format		= VK_FORMAT_R32G32B32A32_SFLOAT;
		vInputAttribDescription0.offset		= offsetof(Particle, pos);

		VkVertexInputAttributeDescription vInputAttribDescription1 {};
		vInputAttribDescription1.location	= 1;
		vInputAttribDescription1.binding	= VERTEX_BUFFER_BIND_ID;
		vInputAttribDescription1.format		= VK_FORMAT_R32G32B32A32_SFLOAT;
		vInputAttribDescription1.offset		= offsetof(Particle, vel);

		// Attribute descriptions
		// Describes memory layout and shader positions
		vertices.attributeDescriptions.resize(2); 
		// Location 0 : Position
		vertices.attributeDescriptions[0]	= vInputAttribDescription0;
		// Location 1 : Velocity
		vertices.attributeDescriptions[1]	= vInputAttribDescription1; 


		// Assign to vertex buffer
		VkPipelineVertexInputStateCreateInfo pipelineVertexInputStateCreateInfo {};
		pipelineVertexInputStateCreateInfo.sType			= VK_STRUCTURE_TYPE_PIPELINE_VERTEX_INPUT_STATE_CREATE_INFO;

		vertices.inputState = pipelineVertexInputStateCreateInfo;
		vertices.inputState.vertexBindingDescriptionCount	= static_cast<uint32_t>(vertices.bindingDescriptions.size());
		vertices.inputState.pVertexBindingDescriptions		= vertices.bindingDescriptions.data();
		vertices.inputState.vertexAttributeDescriptionCount = static_cast<uint32_t>(vertices.attributeDescriptions.size());
		vertices.inputState.pVertexAttributeDescriptions	= vertices.attributeDescriptions.data();
	}




	void SetupDescriptorPool()
	{
		VkDescriptorPoolSize descriptorPoolSize0	= {};
		descriptorPoolSize0.type					= VK_DESCRIPTOR_TYPE_STORAGE_BUFFER;
		descriptorPoolSize0.descriptorCount			= 1;

		std::vector<VkDescriptorPoolSize> poolSizes =
		{
			descriptorPoolSize0,
		};

		VkDescriptorPoolCreateInfo descriptorPoolInfo = {};
		descriptorPoolInfo.sType					= VK_STRUCTURE_TYPE_DESCRIPTOR_POOL_CREATE_INFO;
		descriptorPoolInfo.pNext					= NULL;
		descriptorPoolInfo.poolSizeCount			= poolSizes.size();
		descriptorPoolInfo.pPoolSizes				= poolSizes.data();
		descriptorPoolInfo.maxSets					= 5;


		VK_CHECK_RESULT(vkCreateDescriptorPool(coffeeDevice->device, &descriptorPoolInfo, nullptr, &descriptorPool));
	}


	void SetupDescriptorSetLayout()
	{
		VkDescriptorSetLayoutCreateInfo descriptorLayoutInfo;
		// Shaded layouts (only use first layout binding)
		descriptorLayoutInfo.sType					= VK_STRUCTURE_TYPE_DESCRIPTOR_SET_LAYOUT_CREATE_INFO;
		descriptorLayoutInfo.pNext					= NULL;
		descriptorLayoutInfo.pBindings				= 0;
		descriptorLayoutInfo.bindingCount			= 0;

		VK_CHECK_RESULT(vkCreateDescriptorSetLayout(coffeeDevice->device, 
													&descriptorLayoutInfo, 
													nullptr, 
													&descriptorSetLayout));

	
		VkPipelineLayoutCreateInfo pipelineLayoutInfo = {};
		pipelineLayoutInfo.sType					= VK_STRUCTURE_TYPE_PIPELINE_LAYOUT_CREATE_INFO;
		pipelineLayoutInfo.pNext					= NULL;
		pipelineLayoutInfo.setLayoutCount			= 1;
		pipelineLayoutInfo.pSetLayouts				= &descriptorSetLayout;

		VK_CHECK_RESULT(vkCreatePipelineLayout(coffeeDevice->device, 
											   &pipelineLayoutInfo, 
											   nullptr, 
											   &pipelineLayout));
	}



	void SetupDescriptorSet()
	{
		// Descriptor set
		VkDescriptorSetAllocateInfo allocInfo		= {};
		allocInfo.sType								= VK_STRUCTURE_TYPE_DESCRIPTOR_SET_ALLOCATE_INFO;
		allocInfo.pNext								= NULL;
		allocInfo.descriptorPool					= descriptorPool;
		allocInfo.pSetLayouts						= &descriptorSetLayout;
		allocInfo.descriptorSetCount				= 1;

		VK_CHECK_RESULT(vkAllocateDescriptorSets(coffeeDevice->device, 
												 &allocInfo, 
												 &descriptorSet));
	}


	void SetupComputePipeline()
	{
		// Get a compute `capable' device queue - typically, the graphics and compute
		// queue will be capable - but we need to ask the logical device its abilities

		vkGetDeviceQueue(coffeeDevice->device, 
						 coffeeDevice->queueFamilyIndices.compute, 
						 0, 
						 &coffeeDevice->queue);

		// Create compute pipeline
		// Compute pipelines are created separate from graphics pipelines 
		// even if they use the same queue (family index)

		VkDescriptorSetLayoutBinding setLayoutBinding0 {};
			setLayoutBinding0.descriptorType		= VK_DESCRIPTOR_TYPE_STORAGE_BUFFER;
			setLayoutBinding0.stageFlags			= VK_SHADER_STAGE_COMPUTE_BIT;
			setLayoutBinding0.binding				= 0;
			setLayoutBinding0.descriptorCount		= 1;

		std::vector<VkDescriptorSetLayoutBinding> setLayoutBindings = {
			// Binding 0 : Particle data storage buffer
			setLayoutBinding0,
		};

		VkDescriptorSetLayoutCreateInfo descriptorLayout {};
		descriptorLayout.sType						= VK_STRUCTURE_TYPE_DESCRIPTOR_SET_LAYOUT_CREATE_INFO;
		descriptorLayout.pBindings					= setLayoutBindings.data();
		descriptorLayout.bindingCount				= setLayoutBindings.size();

		VK_CHECK_RESULT(vkCreateDescriptorSetLayout(coffeeDevice->device,	
													&descriptorLayout, 
													nullptr,	
													&compute.descriptorSetLayout));

		VkPipelineLayoutCreateInfo pPipelineLayoutCreateInfo {};
		pPipelineLayoutCreateInfo.sType				= VK_STRUCTURE_TYPE_PIPELINE_LAYOUT_CREATE_INFO;
		pPipelineLayoutCreateInfo.setLayoutCount	= 1;
		pPipelineLayoutCreateInfo.pSetLayouts		= &compute.descriptorSetLayout;

		VK_CHECK_RESULT(vkCreatePipelineLayout(	coffeeDevice->device, 
												&pPipelineLayoutCreateInfo, 
												nullptr,	
												&compute.pipelineLayout));

		VkDescriptorSetAllocateInfo allocInfo {};
		allocInfo.sType								= VK_STRUCTURE_TYPE_DESCRIPTOR_SET_ALLOCATE_INFO;
		allocInfo.descriptorPool					= descriptorPool;
		allocInfo.pSetLayouts						= &compute.descriptorSetLayout;
		allocInfo.descriptorSetCount				= 1;


		VK_CHECK_RESULT(vkAllocateDescriptorSets(coffeeDevice->device, &allocInfo, &compute.descriptorSet));

		VkWriteDescriptorSet writeDescriptorSet0 {};
		writeDescriptorSet0.sType					= VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET;
		writeDescriptorSet0.dstSet					= compute.descriptorSet;
		writeDescriptorSet0.descriptorType			= VK_DESCRIPTOR_TYPE_STORAGE_BUFFER;
		writeDescriptorSet0.dstBinding				= 0;
		writeDescriptorSet0.pBufferInfo				= &compute.storageBuffer.descriptor;
		writeDescriptorSet0.descriptorCount			= 1;

		std::vector<VkWriteDescriptorSet> computeWriteDescriptorSets =
		{
			writeDescriptorSet0,
		};

		vkUpdateDescriptorSets( coffeeDevice->device, 
								static_cast<uint32_t>(computeWriteDescriptorSets.size()), 
								computeWriteDescriptorSets.data(), 
								0, 
								NULL);


		VkComputePipelineCreateInfo computePipelineCreateInfo {};
		computePipelineCreateInfo.sType				= VK_STRUCTURE_TYPE_COMPUTE_PIPELINE_CREATE_INFO;
		computePipelineCreateInfo.layout			= compute.pipelineLayout;
		computePipelineCreateInfo.flags				= 0;

		// Create pipeline		
		computePipelineCreateInfo.stage = LoadShader("model.comp.spv", VK_SHADER_STAGE_COMPUTE_BIT, &shaderModuleCS);

		VK_CHECK_RESULT(vkCreateComputePipelines(coffeeDevice->device, 
						NULL,
						1, 
						&computePipelineCreateInfo, 
						nullptr, 
						&compute.pipeline));

		// Separate command pool as queue family for compute may be different than graphics
		VkCommandPoolCreateInfo cmdPoolInfo			= {};
		cmdPoolInfo.sType							= VK_STRUCTURE_TYPE_COMMAND_POOL_CREATE_INFO;
		cmdPoolInfo.queueFamilyIndex				= coffeeDevice->queueFamilyIndices.compute;
		cmdPoolInfo.flags							= VK_COMMAND_POOL_CREATE_RESET_COMMAND_BUFFER_BIT;

		VK_CHECK_RESULT(vkCreateCommandPool(coffeeDevice->device, 
											&cmdPoolInfo, 
											nullptr, 
											&cmdPool)); // &compute.commandPool));

		VkCommandBufferAllocateInfo cmdBufAllocateInfo {};
		cmdBufAllocateInfo.sType					= VK_STRUCTURE_TYPE_COMMAND_BUFFER_ALLOCATE_INFO;
		cmdBufAllocateInfo.commandPool				= cmdPool; //compute.commandPool;
		cmdBufAllocateInfo.level					= VK_COMMAND_BUFFER_LEVEL_PRIMARY;
		cmdBufAllocateInfo.commandBufferCount		= 1;

		VK_CHECK_RESULT(vkAllocateCommandBuffers(coffeeDevice->device, 
												 &cmdBufAllocateInfo, 
												 &compute.commandBuffer));
	}

	void SetupPipelines()
	{


		VkPipelineInputAssemblyStateCreateInfo inputAssemblyState = {};
		inputAssemblyState.sType					= VK_STRUCTURE_TYPE_PIPELINE_INPUT_ASSEMBLY_STATE_CREATE_INFO;
		inputAssemblyState.topology					= VK_PRIMITIVE_TOPOLOGY_POINT_LIST;
		inputAssemblyState.flags					= 0;
		inputAssemblyState.primitiveRestartEnable	= VK_FALSE;

		VkPipelineRasterizationStateCreateInfo rasterizationState = {};
		rasterizationState.sType					= VK_STRUCTURE_TYPE_PIPELINE_RASTERIZATION_STATE_CREATE_INFO;
		rasterizationState.polygonMode				= VK_POLYGON_MODE_FILL;
		rasterizationState.cullMode					= VK_CULL_MODE_NONE;
		rasterizationState.frontFace				= VK_FRONT_FACE_COUNTER_CLOCKWISE;
		rasterizationState.flags					= 0;
		rasterizationState.depthClampEnable			= VK_FALSE;
		rasterizationState.lineWidth				= 1.0f;

		VkPipelineColorBlendAttachmentState blendAttachmentState = {};
		blendAttachmentState.colorWriteMask			= 0xf;
		blendAttachmentState.blendEnable			= VK_FALSE;

		VkPipelineColorBlendStateCreateInfo colorBlendState = {};
		colorBlendState.sType						= VK_STRUCTURE_TYPE_PIPELINE_COLOR_BLEND_STATE_CREATE_INFO;
		colorBlendState.pNext						= NULL;
		colorBlendState.attachmentCount				= 1;
		colorBlendState.pAttachments				= &blendAttachmentState;

		VkPipelineDepthStencilStateCreateInfo depthStencilState = {};
		depthStencilState.sType						= VK_STRUCTURE_TYPE_PIPELINE_DEPTH_STENCIL_STATE_CREATE_INFO;
		depthStencilState.depthTestEnable			= VK_TRUE;
		depthStencilState.depthWriteEnable			= VK_TRUE;
		depthStencilState.depthCompareOp			= VK_COMPARE_OP_LESS_OR_EQUAL;
		depthStencilState.front						= depthStencilState.back;
		depthStencilState.back.compareOp			= VK_COMPARE_OP_ALWAYS;

		VkPipelineViewportStateCreateInfo viewportState = {};
		viewportState.sType							= VK_STRUCTURE_TYPE_PIPELINE_VIEWPORT_STATE_CREATE_INFO;
		viewportState.viewportCount					= 1;
		viewportState.scissorCount					= 1;
		viewportState.flags							= 0;

		VkPipelineMultisampleStateCreateInfo multisampleState = {};
		multisampleState.sType						= VK_STRUCTURE_TYPE_PIPELINE_MULTISAMPLE_STATE_CREATE_INFO;
		multisampleState.rasterizationSamples		= VK_SAMPLE_COUNT_1_BIT;

		std::vector<VkDynamicState> dynamicStateEnables = 
		{
			VK_DYNAMIC_STATE_VIEWPORT,
			VK_DYNAMIC_STATE_SCISSOR
		};

		VkPipelineDynamicStateCreateInfo dynamicState = {};
		dynamicState.sType							= VK_STRUCTURE_TYPE_PIPELINE_DYNAMIC_STATE_CREATE_INFO;
		dynamicState.pDynamicStates					= dynamicStateEnables.data(),
		dynamicState.dynamicStateCount				= dynamicStateEnables.size();

		// Load shaders
		std::array<VkPipelineShaderStageCreateInfo, 2> shaderStages;

		VkGraphicsPipelineCreateInfo pipelineCreateInfo = {};
		pipelineCreateInfo.sType					= VK_STRUCTURE_TYPE_GRAPHICS_PIPELINE_CREATE_INFO;
		pipelineCreateInfo.pNext					= NULL;
		pipelineCreateInfo.layout					= pipelineLayout; 
		pipelineCreateInfo.renderPass				= renderPass;
		pipelineCreateInfo.flags					= 0;

		pipelineCreateInfo.pVertexInputState		= &vertices.inputState;
		pipelineCreateInfo.pInputAssemblyState		= &inputAssemblyState;
		pipelineCreateInfo.pRasterizationState		= &rasterizationState;
		pipelineCreateInfo.pColorBlendState			= &colorBlendState;
		pipelineCreateInfo.pMultisampleState		= &multisampleState;
		pipelineCreateInfo.pViewportState			= &viewportState;
		pipelineCreateInfo.pDepthStencilState		= &depthStencilState;
		pipelineCreateInfo.pDynamicState			= &dynamicState;
		pipelineCreateInfo.stageCount				= shaderStages.size();
		pipelineCreateInfo.pStages					= shaderStages.data();


		// Scene
		shaderStages[0] = LoadShader("./model.vert.spv", VK_SHADER_STAGE_VERTEX_BIT,   &shaderModuleVS);
		shaderStages[1] = LoadShader("./model.frag.spv", VK_SHADER_STAGE_FRAGMENT_BIT, &shaderModulePS);

		VK_CHECK_RESULT(vkCreateGraphicsPipelines(coffeeDevice->device, 
												  VK_NULL_HANDLE, 
												  1, 
												  &pipelineCreateInfo, 
												  nullptr, 
												  &pipeline));
	}


	void CreateCommandPool()
	{
		VkCommandPoolCreateInfo cmdPoolInfo					= {};
		cmdPoolInfo.sType									= VK_STRUCTURE_TYPE_COMMAND_POOL_CREATE_INFO;
		cmdPoolInfo.queueFamilyIndex						= coffeeDevice->queueFamilyIndices.graphics;//coffeeSwapChain->queueNodeIndex;
		cmdPoolInfo.flags									= VK_COMMAND_POOL_CREATE_RESET_COMMAND_BUFFER_BIT;
		VK_CHECK_RESULT(vkCreateCommandPool(coffeeDevice->device, 
											&cmdPoolInfo, 
											nullptr, 
											&cmdPool));
	}


	void SetupRenderPass()
	{
		// Setup depth stencil pass
		VkImageCreateInfo image								= {};
		image.sType											= VK_STRUCTURE_TYPE_IMAGE_CREATE_INFO;
		image.pNext											= NULL;
		image.imageType										= VK_IMAGE_TYPE_2D;
		image.format										= depthFormat;
		image.extent										= { width, height, 1 };
		image.mipLevels										= 1;
		image.arrayLayers									= 1;
		image.samples										= VK_SAMPLE_COUNT_1_BIT;
		image.tiling										= VK_IMAGE_TILING_OPTIMAL;
		image.usage											= VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT | VK_IMAGE_USAGE_TRANSFER_SRC_BIT;
		image.flags											= 0;

		VkMemoryAllocateInfo mem_alloc						= {};
		mem_alloc.sType										= VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO;
		mem_alloc.pNext										= NULL;
		mem_alloc.allocationSize							= 0;
		mem_alloc.memoryTypeIndex							= 0;

		VkImageViewCreateInfo depthStencilView				= {};
		depthStencilView.sType								= VK_STRUCTURE_TYPE_IMAGE_VIEW_CREATE_INFO;
		depthStencilView.pNext								= NULL;
		depthStencilView.viewType							= VK_IMAGE_VIEW_TYPE_2D;
		depthStencilView.format								= depthFormat;
		depthStencilView.flags								= 0;
		depthStencilView.subresourceRange					= {};
		depthStencilView.subresourceRange.aspectMask		= VK_IMAGE_ASPECT_DEPTH_BIT | VK_IMAGE_ASPECT_STENCIL_BIT;
		depthStencilView.subresourceRange.baseMipLevel		= 0;
		depthStencilView.subresourceRange.levelCount		= 1;
		depthStencilView.subresourceRange.baseArrayLayer	= 0;
		depthStencilView.subresourceRange.layerCount		= 1;

		

		VK_CHECK_RESULT(vkCreateImage(coffeeDevice->device, 
									  &image, 
									  nullptr, 
									  &depthStencil.image));
		
		VkMemoryRequirements memReqs;
		vkGetImageMemoryRequirements(coffeeDevice->device, depthStencil.image, &memReqs);
		mem_alloc.allocationSize			= memReqs.size;
		mem_alloc.memoryTypeIndex			= coffeeDevice->GetMemoryType(memReqs.memoryTypeBits, VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT);
		
		VK_CHECK_RESULT(vkAllocateMemory(coffeeDevice->device, &mem_alloc, nullptr, &depthStencil.mem));
		VK_CHECK_RESULT(vkBindImageMemory(coffeeDevice->device, depthStencil.image, depthStencil.mem, 0));

		depthStencilView.image = depthStencil.image;
		VK_CHECK_RESULT(vkCreateImageView(coffeeDevice->device, &depthStencilView, nullptr, &depthStencil.view));


		std::array<VkAttachmentDescription, 2> attachments = {};
		// Color attachment
		attachments[0].format						= colorformat;
		attachments[0].samples						= VK_SAMPLE_COUNT_1_BIT;
		attachments[0].loadOp						= VK_ATTACHMENT_LOAD_OP_CLEAR;
		attachments[0].storeOp						= VK_ATTACHMENT_STORE_OP_STORE;
		attachments[0].stencilLoadOp				= VK_ATTACHMENT_LOAD_OP_DONT_CARE;
		attachments[0].stencilStoreOp				= VK_ATTACHMENT_STORE_OP_DONT_CARE;
		attachments[0].initialLayout				= VK_IMAGE_LAYOUT_UNDEFINED;
		attachments[0].finalLayout					= VK_IMAGE_LAYOUT_PRESENT_SRC_KHR;
		// Depth attachment
		attachments[1].format						= depthFormat;
		attachments[1].samples						= VK_SAMPLE_COUNT_1_BIT;
		attachments[1].loadOp						= VK_ATTACHMENT_LOAD_OP_CLEAR;
		attachments[1].storeOp						= VK_ATTACHMENT_STORE_OP_STORE;
		attachments[1].stencilLoadOp				= VK_ATTACHMENT_LOAD_OP_DONT_CARE;
		attachments[1].stencilStoreOp				= VK_ATTACHMENT_STORE_OP_DONT_CARE;
		attachments[1].initialLayout				= VK_IMAGE_LAYOUT_UNDEFINED;
		attachments[1].finalLayout					= VK_IMAGE_LAYOUT_DEPTH_STENCIL_ATTACHMENT_OPTIMAL;

		VkAttachmentReference colorReference		= {};
		colorReference.attachment					= 0;
		colorReference.layout						= VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL;

		VkAttachmentReference depthReference		= {};
		depthReference.attachment					= 1;
		depthReference.layout						= VK_IMAGE_LAYOUT_DEPTH_STENCIL_ATTACHMENT_OPTIMAL;

		VkSubpassDescription subpassDescription		= {};
		subpassDescription.pipelineBindPoint		= VK_PIPELINE_BIND_POINT_GRAPHICS;
		subpassDescription.colorAttachmentCount		= 1;
		subpassDescription.pColorAttachments		= &colorReference;
		subpassDescription.pDepthStencilAttachment	= &depthReference;
		subpassDescription.inputAttachmentCount		= 0;
		subpassDescription.pInputAttachments		= nullptr;
		subpassDescription.preserveAttachmentCount	= 0;
		subpassDescription.pPreserveAttachments		= nullptr;
		subpassDescription.pResolveAttachments		= nullptr;

		// Subpass dependencies for layout transitions
		std::array<VkSubpassDependency, 2> dependencies;

		dependencies[0].srcSubpass					= VK_SUBPASS_EXTERNAL;
		dependencies[0].dstSubpass					= 0;
		dependencies[0].srcStageMask				= VK_PIPELINE_STAGE_BOTTOM_OF_PIPE_BIT;
		dependencies[0].dstStageMask				= VK_PIPELINE_STAGE_COLOR_ATTACHMENT_OUTPUT_BIT;
		dependencies[0].srcAccessMask				= VK_ACCESS_MEMORY_READ_BIT;
		dependencies[0].dstAccessMask				= VK_ACCESS_COLOR_ATTACHMENT_READ_BIT | VK_ACCESS_COLOR_ATTACHMENT_WRITE_BIT;
		dependencies[0].dependencyFlags				= VK_DEPENDENCY_BY_REGION_BIT;

		dependencies[1].srcSubpass					= 0;
		dependencies[1].dstSubpass					= VK_SUBPASS_EXTERNAL;
		dependencies[1].srcStageMask				= VK_PIPELINE_STAGE_COLOR_ATTACHMENT_OUTPUT_BIT;
		dependencies[1].dstStageMask				= VK_PIPELINE_STAGE_BOTTOM_OF_PIPE_BIT;
		dependencies[1].srcAccessMask				= VK_ACCESS_COLOR_ATTACHMENT_READ_BIT | VK_ACCESS_COLOR_ATTACHMENT_WRITE_BIT;
		dependencies[1].dstAccessMask				= VK_ACCESS_MEMORY_READ_BIT;
		dependencies[1].dependencyFlags				= VK_DEPENDENCY_BY_REGION_BIT;

		VkRenderPassCreateInfo renderPassInfo		= {};
		renderPassInfo.sType						= VK_STRUCTURE_TYPE_RENDER_PASS_CREATE_INFO;
		renderPassInfo.attachmentCount				= static_cast<uint32_t>(attachments.size());
		renderPassInfo.pAttachments					= attachments.data();
		renderPassInfo.subpassCount					= 1;
		renderPassInfo.pSubpasses					= &subpassDescription;
		renderPassInfo.dependencyCount				= static_cast<uint32_t>(dependencies.size());
		renderPassInfo.pDependencies				= dependencies.data();

		VK_CHECK_RESULT(vkCreateRenderPass( coffeeDevice->device, 
											&renderPassInfo, 
											nullptr, 
											&renderPass));
	}

	

	
	void SetupFrameBuffer()
	{
		VkImageView attachments[2];

		// Depth/Stencil attachment is the same for all frame buffers
		attachments[1] = depthStencil.view;

		VkFramebufferCreateInfo frameBufferCreateInfo = {};
		frameBufferCreateInfo.sType					= VK_STRUCTURE_TYPE_FRAMEBUFFER_CREATE_INFO;
		frameBufferCreateInfo.pNext					= NULL;
		frameBufferCreateInfo.renderPass			= renderPass;
		frameBufferCreateInfo.attachmentCount		= 2;
		frameBufferCreateInfo.pAttachments			= attachments;
		frameBufferCreateInfo.width					= width;
		frameBufferCreateInfo.height				= height;
		frameBufferCreateInfo.layers				= 1;

		// Create frame buffers for every swap chain image
		frameBuffers.resize(coffeeSwapChain->imageCount);
		for (uint32_t i = 0; i < frameBuffers.size(); i++)
		{
			attachments[0] = coffeeSwapChain->imageViews[i];
			VK_CHECK_RESULT(vkCreateFramebuffer(coffeeDevice->device, &frameBufferCreateInfo, nullptr, &frameBuffers[i]));
		}
	}


	void SetupVulkan()
	{
		// Vulkan device creation
		DBG_ASSERT(coffeeDevice==NULL);
		coffeeDevice = new VkCoffeeDevice();

		DBG_ASSERT(coffeeSwapChain==NULL);
		coffeeSwapChain = new VkCoffeeSwapChain(coffeeDevice->instance, 
												coffeeDevice->device,
												coffeeDevice->physicalDevice,
												windowHWND);
		width  = coffeeSwapChain->width;
		height = coffeeSwapChain->height;

		CreateCommandPool();
		CreateCommandBuffers();
		/*
			call `BuildCommandBuffer' to update the commands we want to perform
			such as drawing and vertices
		*/

		SetupRenderPass();
		SetupFrameBuffer();

		SetupDescriptorPool();
		SetupDescriptorSetLayout();
		SetupDescriptorSet();

		// `compute' pipeline/buffers (buffer shared between graphics and compute)
		CreateComputeStorageBuffers();
		SetupComputePipeline();


		SetupPipelines();


	}


	

	void BuildCommandBuffers()
	{
		VkCommandBufferBeginInfo cmdBufInfo			= {};
		cmdBufInfo.sType							= VK_STRUCTURE_TYPE_COMMAND_BUFFER_BEGIN_INFO;
		cmdBufInfo.pNext							= NULL;

		VkClearValue clearValues[2];
		clearValues[0].color						= { { 0.025f, 0.025f, 0.225f, 1.0f } };
		clearValues[1].depthStencil					= { 1.0f, 0 };

		VkRenderPassBeginInfo renderPassBeginInfo	= {};
		renderPassBeginInfo.sType                   = VK_STRUCTURE_TYPE_RENDER_PASS_BEGIN_INFO;
		renderPassBeginInfo.pNext                   = NULL;
		renderPassBeginInfo.renderPass              = renderPass;
		renderPassBeginInfo.renderArea.offset.x     = 0;
		renderPassBeginInfo.renderArea.offset.y     = 0;
		renderPassBeginInfo.renderArea.extent.width = width;
		renderPassBeginInfo.renderArea.extent.height= height;
		renderPassBeginInfo.clearValueCount         = 2;
		renderPassBeginInfo.pClearValues            = clearValues;

		for (int32_t i = 0; i < 2; ++i)
		{
			// Set target frame buffer
			renderPassBeginInfo.framebuffer = frameBuffers[i]; 

			VK_CHECK_RESULT(vkBeginCommandBuffer(clearCmdBuffers[i], &cmdBufInfo));

			vkCmdBeginRenderPass(clearCmdBuffers[i], &renderPassBeginInfo, VK_SUBPASS_CONTENTS_INLINE);

			VkViewport viewport = {};
			viewport.width    = (float)width;
			viewport.height   = (float)height;
			viewport.minDepth = 0;
			viewport.maxDepth = 1;
			vkCmdSetViewport(clearCmdBuffers[i], 0, 1, &viewport);

			VkRect2D scissor      = {};
			scissor.extent.width  = width;
			scissor.extent.height = height;
			scissor.offset.x      = 0;
			scissor.offset.y      = 0;
			vkCmdSetScissor(clearCmdBuffers[i], 0, 1, &scissor);

			VkDeviceSize offsets[1] = { 0 };
			
			// Draw particles
			{
			vkCmdBindDescriptorSets(clearCmdBuffers[i], VK_PIPELINE_BIND_POINT_GRAPHICS, 
									pipelineLayout, // graphics.pipelineLayout, 
									0, 1, 
									&descriptorSet, // &graphics.descriptorSet, 
									0, NULL);
			vkCmdBindPipeline(clearCmdBuffers[i], VK_PIPELINE_BIND_POINT_GRAPHICS, pipeline);//graphics.pipeline);
			
			vkCmdBindVertexBuffers(clearCmdBuffers[i], VERTEX_BUFFER_BIND_ID, 1, &compute.storageBuffer.buffer, offsets);
			vkCmdDraw(clearCmdBuffers[i], PARTICLE_COUNT, 1, 0, 0);
			}

			vkCmdEndRenderPass(clearCmdBuffers[i]);

			VK_CHECK_RESULT(vkEndCommandBuffer(clearCmdBuffers[i]));
		}
	}

	void Render()
	{
		if (!coffeeSwapChain)
			return;

		
		VK_CHECK_RESULT(coffeeSwapChain->AcquireNextImage(coffeeDevice->device, 
		                                                  NULL, 
												          &currentBuffer));
		static float delay    = 0;
		delay				+= 0.002f;

		char buf[256] = {0};
		sprintf(buf, "%.2f, delay", delay);
		::SetWindowText(windowHWND, buf); 

		if ( delay > 1 )
		{
			delay = 0;
		}


		BuildComputeCommandBuffer();

		BuildCommandBuffers();

		// **1** Rendering to the screen
		// Clear screen and draw particle points (graphics cmds)
		{
		VkFence           renderFence;
		VkFenceCreateInfo fenceCreateInfo = {};
		fenceCreateInfo.sType = VK_STRUCTURE_TYPE_FENCE_CREATE_INFO;
		vkCreateFence( coffeeDevice->device, &fenceCreateInfo, NULL, &renderFence );

		VkSubmitInfo si = {};
		si.sType                = VK_STRUCTURE_TYPE_SUBMIT_INFO;
		si.waitSemaphoreCount   = 0;
		si.pWaitSemaphores      = VK_NULL_HANDLE;
		si.pWaitDstStageMask    = NULL;
		si.commandBufferCount   = 1;
		si.pCommandBuffers      = &clearCmdBuffers[currentBuffer];
		si.signalSemaphoreCount = 0;
		si.pSignalSemaphores    = VK_NULL_HANDLE;

		vkQueueSubmit( coffeeDevice->queue, 1, &si, renderFence );

		vkWaitForFences( coffeeDevice->device, 1, &renderFence, VK_TRUE, UINT64_MAX );
		vkDestroyFence( coffeeDevice->device, renderFence, NULL );
		}
		VK_CHECK_RESULT(vkQueueWaitIdle(coffeeDevice->queue));

		VK_CHECK_RESULT(coffeeSwapChain->QueuePresent(coffeeDevice->queue, 
		                                              currentBuffer, 
													  NULL));

		
		// Wait for the queue to become idle
		VK_CHECK_RESULT(vkQueueWaitIdle(coffeeDevice->queue));

		// ** 2 ** Compute pass
		{
			VkFence           renderFence;
			VkFenceCreateInfo fenceCreateInfo = {};
			fenceCreateInfo.sType = VK_STRUCTURE_TYPE_FENCE_CREATE_INFO;
			vkCreateFence( coffeeDevice->device, &fenceCreateInfo, NULL, &renderFence );

			VkSubmitInfo si = {};
			si.sType                = VK_STRUCTURE_TYPE_SUBMIT_INFO;
			si.commandBufferCount   = 1;
			si.pCommandBuffers      = &compute.commandBuffer;

			VK_CHECK_RESULT(vkQueueSubmit(coffeeDevice->queue, 1, &si, renderFence));

			vkWaitForFences( coffeeDevice->device, 1, &renderFence, VK_TRUE, UINT64_MAX );
			vkDestroyFence( coffeeDevice->device, renderFence, NULL );
		}

		// Wait for the queue to become idle (avoid compute/graphics conflicting)
		VK_CHECK_RESULT(vkQueueWaitIdle(coffeeDevice->queue));

	}

	void UpdateUniformBuffers()
	{
		//ubo.model      = Matrix4::SetRotationAxis(-rotation.y, Vector3(0,1,0)) *
		//			     Matrix4::SetRotationAxis(-rotation.x, Vector3(1,0,0));

		//ubo.projection = Matrix4::SetPerspectiveRH( 1.6f,//45, radians not degrees
		//									        (float)width / (float)height,
		//									        2.5f, 
		//									        400.0f);

		//ubo.view       = Matrix4::SetLookAtRH( Vector3(0,4,-4), 
		//								       Vector3(0,0,0),
		//								       Vector3(0,1,0));

		#if 0
		uint8_t *pData;
		VK_CHECK_RESULT(vkMapMemory(coffeeDevice->device, uniformData.memory, 0, sizeof(ubo), 0, (void **)&pData));
		memcpy(pData, &ubo, sizeof(ubo));
		vkUnmapMemory(coffeeDevice->device, uniformData.memory);
		#endif
	}





	HWND SetupWindow(HINSTANCE hinstance, WNDPROC wndproc)
	{
		this->windowInstance	= hinstance;

		WNDCLASSEX wndClass;
		wndClass.cbSize         = sizeof(WNDCLASSEX);
		wndClass.style          = CS_HREDRAW | CS_VREDRAW;
		wndClass.lpfnWndProc    = wndproc;
		wndClass.cbClsExtra     = 0;
		wndClass.cbWndExtra     = 0;
		wndClass.hInstance      = hinstance;
		wndClass.hIcon          = LoadIcon(NULL, IDI_APPLICATION);
		wndClass.hCursor        = LoadCursor(NULL, IDC_ARROW);
		wndClass.hbrBackground  = (HBRUSH)GetStockObject(BLACK_BRUSH);
		wndClass.lpszMenuName   = NULL;
		wndClass.lpszClassName  = name;
		wndClass.hIconSm        = LoadIcon(NULL, IDI_WINLOGO);

		if (!RegisterClassEx(&wndClass))
		{
			DBG_ASSERT_MSG(false, "Could not register window class!");
		}

		DWORD dwExStyle   = WS_EX_APPWINDOW | WS_EX_WINDOWEDGE;
		DWORD dwStyle     = WS_OVERLAPPEDWINDOW | WS_CLIPSIBLINGS | WS_CLIPCHILDREN;

		RECT windowRect;
		windowRect.left   = 0L;
		windowRect.top    = 0L;
		windowRect.right  = (long)width;
		windowRect.bottom = (long)height;

		AdjustWindowRectEx(&windowRect, dwStyle, FALSE, dwExStyle);

		std::string windowTitle = name;
		windowHWND = CreateWindowEx(0,
			name,
			windowTitle.c_str(),
			dwStyle | WS_CLIPSIBLINGS | WS_CLIPCHILDREN,
			0,
			0,
			windowRect.right - windowRect.left,
			windowRect.bottom - windowRect.top,
			NULL,
			NULL,
			hinstance,
			NULL);


		// Center on screen
		uint32_t x = (GetSystemMetrics(SM_CXSCREEN) - windowRect.right) / 2;
		uint32_t y = (GetSystemMetrics(SM_CYSCREEN) - windowRect.bottom) / 2;
		SetWindowPos(windowHWND, 0, x, y, 0, 0, SWP_NOZORDER | SWP_NOSIZE);

		DBG_ASSERT_MSG(windowHWND, "Could not create window!");

		ShowWindow(windowHWND, SW_SHOW);
		SetForegroundWindow(windowHWND);
		SetFocus(windowHWND);

		return windowHWND;
	}


	void HandleMessages(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
	{
		switch (uMsg)
		{
			case WM_CLOSE: 
			{
				// posts a WM_QUIT message
				PostQuitMessage( 0 );
			}
			break;

			case WM_PAINT:
				ValidateRect(windowHWND, NULL);
			break;

			case WM_KEYDOWN:
			break;

			case WM_RBUTTONDOWN:
			case WM_LBUTTONDOWN:
			case WM_MBUTTONDOWN:
			break;

			case WM_MOUSEWHEEL:
			{
				short wheelDelta = GET_WHEEL_DELTA_WPARAM(wParam);
			}
			break;

			case WM_MOUSEMOVE:
			if (wParam & MK_RBUTTON)
			{
				int32_t posx = LOWORD(lParam);
				int32_t posy = HIWORD(lParam);
			}
			if (wParam & MK_LBUTTON)
			{
				int32_t posx = LOWORD(lParam);
				int32_t posy = HIWORD(lParam);
			}
			if (wParam & MK_MBUTTON)
			{
				int32_t posx = LOWORD(lParam);
				int32_t posy = HIWORD(lParam);
			}
			break;
		}
	}	


	void StartRenderLoop()
	{
		MSG msg;
		while (TRUE)
		{
			while (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
			{
				TranslateMessage(&msg);
				DispatchMessage(&msg);
			}

			if (msg.message == WM_QUIT)
			{
				break;
			}

			Render();
		}

		// Flush coffeeDevice->device to make sure all resources can be freed 
		vkDeviceWaitIdle(coffeeDevice->device);
	}// End StartRenderLoop(..)
};




static
VulkanDemo *vulkanDemo = NULL;				
														
LRESULT CALLBACK WndProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)						
{																									
	if (vulkanDemo != NULL)																		
	{																								
		vulkanDemo->HandleMessages(hWnd, uMsg, wParam, lParam);									
	}																								
	return (DefWindowProc(hWnd, uMsg, wParam, lParam));												
}										

		
// Program Entry Point													
int APIENTRY WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR pCmdLine, int nCmdShow)	
{																									
	vulkanDemo = new VulkanDemo();															
	vulkanDemo->SetupWindow(hInstance, WndProc);																														
	vulkanDemo->SetupVulkan();																		
	vulkanDemo->StartRenderLoop();																	
	delete(vulkanDemo);																			
	return 0;																						
}// End WinMain(..)	