

#pragma once

#include <vector>

#include "vulkan.h"

struct Buffer
{
	VkBuffer				buffer				= 0;
	VkDeviceMemory			memory				= 0;
	VkDescriptorBufferInfo	descriptor;
	
	VkDeviceSize			alignment			= 0;
	VkDeviceSize			size				= 0;
	void*					mapped				= nullptr;
	VkBufferUsageFlags		usageFlags			= 0;
	VkMemoryPropertyFlags	memoryPropertyFlags	= 0;

	static
	uint32_t GetMemoryType(uint32_t					typeBits, 
	                       VkMemoryPropertyFlags	properties, 
						   VkBool32 *				memTypeFound = nullptr)
	{
		extern VkCoffeeDevice *				coffeeDevice;

		// Read the device memory properties
		VkPhysicalDeviceMemoryProperties memoryProperties;
		vkGetPhysicalDeviceMemoryProperties( coffeeDevice->physicalDevice, &memoryProperties );

		for (uint32_t i = 0; i < memoryProperties.memoryTypeCount; i++)
		{
			if ((typeBits & 1) == 1)
			{
				if ((memoryProperties.memoryTypes[i].propertyFlags & properties) == properties)
				{
					if (memTypeFound)
					{
						*memTypeFound = true;
					}
					return i;
				}
			}
			typeBits >>= 1;
		}

		if (memTypeFound)
		{
			*memTypeFound = false;
			return 0;
		}
		
		__debugbreak();
		throw std::runtime_error("Could not find a matching memory type");
		return 0;
	}

	static
	VkResult CreateBuffer(VkBufferUsageFlags	usageFlags, 
						  VkMemoryPropertyFlags memoryPropertyFlags, 
						  Buffer				*buffer, 
						  VkDeviceSize			size, 
						  void					*data = nullptr)
	{
		extern VkCoffeeDevice *				coffeeDevice;

		// Create the buffer handle
		VkBufferCreateInfo bufferCreateInfo {};
		bufferCreateInfo.sType = VK_STRUCTURE_TYPE_BUFFER_CREATE_INFO;
		bufferCreateInfo.usage = usageFlags;
		bufferCreateInfo.size = size;
		VK_CHECK_RESULT(vkCreateBuffer(coffeeDevice->device, &bufferCreateInfo, nullptr, &buffer->buffer));

		// Create the memory backing up the buffer handle
		VkMemoryRequirements memReqs;
		VkMemoryAllocateInfo memAlloc {};
		memAlloc.sType = VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO;
		vkGetBufferMemoryRequirements(coffeeDevice->device, buffer->buffer, &memReqs);
		memAlloc.allocationSize = memReqs.size;
		// Find a memory type index that fits the properties of the buffer
		memAlloc.memoryTypeIndex = GetMemoryType(memReqs.memoryTypeBits, memoryPropertyFlags);
		VK_CHECK_RESULT(vkAllocateMemory(coffeeDevice->device, &memAlloc, nullptr, &buffer->memory));

		buffer->alignment			= memReqs.alignment;
		buffer->size				= memAlloc.allocationSize;
		buffer->usageFlags			= usageFlags;
		buffer->memoryPropertyFlags = memoryPropertyFlags;

		// If a pointer to the buffer data has been passed, map the buffer and copy over the data
		if (data != nullptr)
		{
			vkMapMemory(coffeeDevice->device, 
						buffer->memory, 
						0, 
						size, 
						0, 
						&buffer->mapped);

			memcpy(buffer->mapped, data, (size_t)size);

			vkUnmapMemory(coffeeDevice->device, buffer->memory);
			buffer->mapped = nullptr;
		}

		// Initialize a default descriptor that covers the whole buffer size
		buffer->descriptor.offset	= 0;
		buffer->descriptor.buffer	= buffer->buffer;
		buffer->descriptor.range	= VK_WHOLE_SIZE;

		// Attach the memory to the buffer object
		return vkBindBufferMemory(coffeeDevice->device, buffer->buffer, buffer->memory, 0);
	}

};
